/*
RetroEngine Test Program

This is a quick and dirty test program that uses several 
RetroEngine features. I would not recommend following this
as a template for writing a game (though honestly it would
probably be more in-keeping with slapdash 1990s practices).

*/
#include <iostream>
#include <math.h>
#include <string.h>
#include "SDL2/SDL.h"
#include "sdl_virtual_screen.h"
#include "RetroEngine.h"

#ifdef WIN32
// create a usleep() function that matches the unix standard
// so that we don't need to adjust code elsewhere
#include <Windows.h>
#define usleep(a) Sleep((a)/1000 + 1)
#else
// include unistd.h for usleep()
#include <unistd.h>
#endif

using namespace std;

/**
 * A way to simulate waiting for the old notion of waiting for a 
 * vsync interrupt. Designed to release the CPU as much as possible
 * while waiting.
 */
class CPacemaker {
public:
	Uint64 interval;
	Uint64 threshhold;
	void wait() {
		// figure out when the next "vblank" will be (the next even multiple of our interval)
		Uint64 nextBeat = (SDL_GetPerformanceCounter()/this->interval + 1)*this->interval;

		// if we're far enough away from the next frame, relinquish the CPU
		while ((SDL_GetPerformanceCounter() < nextBeat)
			&& (nextBeat - SDL_GetPerformanceCounter() > threshhold)) {
			usleep(999);
		} 

		// if we're REALLY CLOSE to the next frame, enter a busy wait
		while (SDL_GetPerformanceCounter() < nextBeat) {
		} 
	}

	void setThreshhold(Uint64 inMilliseconds) {
		this->threshhold = inMilliseconds * SDL_GetPerformanceFrequency()/1000;
	}

	CPacemaker():threshhold(1000) {};
};


// GLOBAL CONSTANTS - resolution settings
const bool FullScreen = false;
const int InternalScreenWidth = 480;
const int InternalScreenHeight = 270;
const int WindowedScale = 3;
const int DefaultScreenWidth = InternalScreenWidth * WindowedScale;
const int DefaultScreenHeight = InternalScreenHeight * WindowedScale;

// GLOBAL VARIABLES
bool gameRunning = true;
SDL_Window *win;
Uint32 currentFrame;

// GLOBAL FUNCTIONS - forward declarations
void InitGame();
void FinishGame();
void MyHBlank(CRetroEngine *inRe, int scanline, void *payload);
void LoadChrMemory(uint8* chr_memory);
void DecompressSNES(uint8 *chr_memory);
void LoadPal(uint32 *pal);

int main (int argc, char *argv[]) {
	// ---------------------------------
	// DATA


	// timing data - for frame pacing and FPS measurement
	CPacemaker pacemaker;

	// some tables for reducing in-frame calculations
	int sintab[256];
	int costab[256];

	// string for outputting FPS
	char FrameTitle[1024];

	// event object
	SDL_Event evt;

	// loop conters, frame counter, array iterator
	int x, y, frameNo=0;
	unsigned char *offs;
	CSDLVirtualScreen csvsScreen;
	CRetroEngine re;



	// ----------------------------------
	// INITIALIZATION

	// Init SDL, our output buffer, and RetroEngine
	InitGame();
	csvsScreen.Initialize(InternalScreenWidth, InternalScreenHeight, win);
	re.SetResolution(InternalScreenWidth, InternalScreenHeight);

	// fill char memory
	LoadChrMemory(re.chr_memory);
	DecompressSNES(re.chr_memory);

	// fill palette memory
	LoadPal(re.palette);

	// set timer
	pacemaker.interval = SDL_GetPerformanceFrequency()/60;
#ifdef WIN32
	// windows makes the app sleep for too long, so let's
	// set the threshhold at > 1 frame of animation (16 ms)
	pacemaker.setThreshhold(20);
#else 
	// If we're more than 10 ms away from the next frame, we'll
	// take a quick sleep. Once we get within 10 ms of the frame,
	// we'll enter a busy wait.
	pacemaker.setThreshhold(10);
#endif

	// fill cosine and sine tabs
	for (y=0; y<256; y++) {
		costab[y] = (int)(InternalScreenHeight*0.75f*0.5f*cos((float)y*2.0f*3.14159f/256.0f));
		sintab[y] = (int)(InternalScreenHeight*0.75f*0.5f*sin((float)y*2.0f*3.14159f/256.0f));
	}

	// Initialize tilemaps
	re.SetLayer(3, true, 0, 0, RE_BLEND_OFF, 0);
	re.SetLayer(2, true, 0, 0, RE_BLEND_OFF, 0);
	for (y=0;y<RE_TILEMAP_SIZE_Y;y++) {
		for (x=0; x<RE_TILEMAP_SIZE_X; x++) {
			// Use tile flipping to put a pattern in layer 3
			re.tilemap[3][y*RE_TILEMAP_SIZE_Y + x] = re_make_tile(0x10, 0, x&0x0001, y&0x0001, 1);
			// put blank tiles in layer 2
			re.tilemap[2][y*RE_TILEMAP_SIZE_Y + x] = re_make_tile(0x00, 0, false, false, 1);
		}
	}

	// ---------------------------
	// GAME LOOP

	// enter game loop
	while (gameRunning) {
		// process the SDL event queue
		while (SDL_PollEvent(&evt)) {
			// close button
			if (evt.type == SDL_QUIT) {
				gameRunning = false;
			}
			// use ESC to close also
			if (evt.type == SDL_KEYDOWN) {
				switch(evt.key.keysym.sym) {
					case SDLK_ESCAPE: 
						gameRunning = false;
						break;
				}
			}
		}

		// set our HBlank 
		re.SetHBlank(MyHBlank, nullptr);

		// scroll layer 3
		re.SetLayer(3, true, frameNo, frameNo >> 1, RE_BLEND_OFF, 0);

		// create ball sprites
		y=0;
		for (x=8; x<InternalScreenWidth; x+=32) {
			re.SetSprite(y++, true, 0x01, 0, 0, 2, 0, x, (InternalScreenHeight>>1) - sintab[((x>>2) + frameNo*2)&0xFF] - 4, false, false, 0);
			re.SetSprite(y++, true, 0x01, 0, 0, 2, 0, x+16, (InternalScreenHeight>>1) + sintab[((x>>2) + frameNo*2)&0xFF] - 4, false, false, 0);
		}

		// create logo sprites
		re.SetSprite(y++, true, 0x02, 3, 1, 2, 1, InternalScreenWidth-80, InternalScreenHeight-16, false, false, 0);
		re.SetSprite(y++, true, 0x06, 3, 1, 2, 1, InternalScreenWidth-80+32, InternalScreenHeight-16, false, false, 0);
		re.SetSprite(y++, true, 0x0A, 0, 1, 2, 1, InternalScreenWidth-80+64, InternalScreenHeight-16, false, false, 0);

		// draw screen and copy to window
		re.Draw();

		// Copy RetroEngine output to vitual screen
		csvsScreen.DirectUpdate(re.GetScreen());

		// wait for simulated VBLANK
		pacemaker.wait();

		// Execute SDL Blit. -- This is where the virtual screen actually gets rendered to the window.
		//						We don't want ANYTHING between our pacemaker and blit.
		csvsScreen.RenderTextureToScreen();

		// increment frame number (for animating layers)
		frameNo++;

	}

	// shut down SDL
	FinishGame();


	return 0;
}

/**
Initializes SDL window according to global constants.
*/
void InitGame() {
	SDL_Init(SDL_INIT_EVERYTHING);
	char WindowTitle[64] = "RetroEngine Test";

	if (FullScreen) {
		// When creating a fullscreen window, don't change the current desktop mode. Just make a big window.
		win = SDL_CreateWindow(WindowTitle, SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, 0, 0, SDL_WINDOW_FULLSCREEN_DESKTOP);
		SDL_ShowCursor(SDL_FALSE);	// hide cursor when fullscreen
	} else {
		win = SDL_CreateWindow(WindowTitle, SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, DefaultScreenWidth, DefaultScreenHeight, SDL_WINDOW_SHOWN);
	}
}

/**
Shuts down SDL and exits
*/
void FinishGame() {
	SDL_DestroyWindow(win);
	SDL_Quit();
	exit(0); // should I really do this here? Should I return and exit?
}

/**
A callback for each line of the rendered image
*/
void MyHBlank(CRetroEngine *inRE, int scanline, void *payload) {
	int top = InternalScreenHeight >> 2;
	int bottom = (top << 1) + top;
	re_layer tLayer = inRE->GetLayer(3);

	if ((scanline > top) && (scanline < bottom)) {
		tLayer |= RE_LAYER_ENABLE_FLAG;
	} else {
		tLayer &= ~RE_LAYER_ENABLE_FLAG;
	}

	inRE->SetLayerRaw(3, tLayer);
};


/**
Loads binary sprites into chr memory
*/
void LoadChrMemory(uint8* chr_memory) {
	FILE *fp;
	size_t chunkSize = 8*8*1024/2;
	size_t bytesRead;

	// we have a relative path, but let's get an absolute path
	// using a cross-platform method
	char *data_path = SDL_GetBasePath();
	char filename[2048] = "";
	if (strstr(data_path, "\\Release\\")!=NULL || strstr(data_path, "\\Debug\\")!=NULL) {
		// We're running from inside the Visual Studio IDE, so do nothing -
		// let the Visual Studio project resolve relative paths.
	} else {
		// we're not running from inside Visual Studio, so figure out where we are
		strcat(filename, data_path);
	}
	strcat(filename, "assets/demo.chr");

	// open file
	fp = fopen(filename, "rb");

	if (fp) {
		// read it all in one go
		bytesRead = fread(chr_memory, sizeof(uint8), chunkSize, fp);
		fclose(fp);
	}

}
/**
 * Loads the palette and converts to ARGB for 32-bit systems
 */
void LoadPal(uint32 *pal) {
	FILE *fp;
	size_t bytesRead;
	uint8 *inputBuffer, *tib;
	int i;

	// convert relative path to absolute path
	char *data_path = SDL_GetBasePath();
	char filename[2048] = "";
	if (strstr(data_path, "\\Release\\")!=NULL || strstr(data_path, "\\Debug\\")!=NULL) {
		// We're running from inside the Visual Studio IDE, so do nothing -
		// let the Visual Studio project resolve relative paths.
	} else {
		// we're not running from inside Visual Studio, so figure out where we are
		strcat(filename, data_path);
	}
	strcat(filename, "assets/demo.pal");

	// open file
	fp = fopen(filename, "rb");

	if (fp) {
		// create input buffer and read in the palette as 768 bytes
		inputBuffer = (uint8 *)malloc(768);
		bytesRead = fread(inputBuffer, sizeof(uint8), 768, fp);

		// get a pointer to the front of the buffer
		tib = inputBuffer;

		// iterate through all 256 colors
		for (i=0;i<256;i++) {

			// convert 3 bytes into a single 32-bit unsigned integer
			*pal = tib[0]<<16 | tib[1]<<8 | tib[2];

			// advance to next color
			pal++;
			tib += 3;
		}

		// clean up
		free(inputBuffer);
		fclose(fp);
	}
}

/**
 * Converts SNES bit-planes into simple discrete values
 */
void DecompressSNES(uint8 *chr_memory) {
	uint8 *tmp_memory;
	uint8 *readptr, *writeptr;
	uint32 b0, b1, b2, b3;
	int i, j;

	// allocate temporary storage
	tmp_memory = (uint8*)malloc(32768);
	for (i = 0; i < 1024; i++) {
		// go to the sprite index in each map
		readptr = chr_memory + 32*i;
		writeptr = tmp_memory + ((i&0x0f)<<2) + ((i>>4)<<9);
		for (j = 0; j < 8; j++) {
			// get bitplane 0 & 1
			b0 = readptr[0];
			b1 = readptr[1];
			// de-leave planes 0 and 1
			b0 = (b0 | (b0<<4)) & 0x0f0f;
			b0 = (b0 | (b0<<2)) & 0x3333;
			b0 = (b0 | (b0<<1)) & 0x5555;
			b1 = (b1 | (b1<<4)) & 0x0f0f;
			b1 = (b1 | (b1<<2)) & 0x3333;
			b1 = (b1 | (b1<<1)) & 0x5555;
			b0 = b0 | (b1<<1);
			// get bitplane 2 & 3
			b2 = readptr[16];
			b3 = readptr[17];
			// de-leave planes 2 and 3
			b2 = (b2 | (b2<<4)) & 0x0f0f;
			b2 = (b2 | (b2<<2)) & 0x3333;
			b2 = (b2 | (b2<<1)) & 0x5555;
			b3 = (b3 | (b3<<4)) & 0x0f0f;
			b3 = (b3 | (b3<<2)) & 0x3333;
			b3 = (b3 | (b3<<1)) & 0x5555;
			b2 = b2 | (b3<<1);
			// combine as: 22002200220022002200220022002200
			b0 = (b0 | (b0<<8)) & 0x00ff00ff00ff00ff;
			b0 = (b0 | (b0<<4)) & 0x0f0f0f0f0f0f0f0f;
			b0 = (b0 | (b0<<2)) & 0x3333333333333333;
			b2 = (b2 | (b2<<8)) & 0x00ff00ff00ff00ff;
			b2 = (b2 | (b2<<4)) & 0x0f0f0f0f0f0f0f0f;
			b2 = (b2 | (b2<<2)) & 0x3333333333333333;
			b0 = b0 | (b2<<2);
			// save data (this is clumsy but we can't guarantee endian-ness)
			writeptr[0] = (b0>>24) & 0xff;
			writeptr[1] = (b0>>16) & 0xff;
			writeptr[2] = (b0>>8)  & 0xff;
			writeptr[3] = (b0>>0)  & 0xff;
			// advance to next
			readptr += 2;
			writeptr += 64;
		}
	}

	// copy memory
	memcpy(chr_memory, tmp_memory, 32768);

	// free temp memory
	free(tmp_memory);
}
