#include <stdio.h>
#include <string.h>
#include "SDL2/SDL.h"
#include "sdl_virtual_screen.h"

CSDLVirtualScreen::~CSDLVirtualScreen() {
	if (this->ren != nullptr) SDL_DestroyRenderer(this->ren);
}

int CSDLVirtualScreen::Initialize(int inWidth, int inHeight, SDL_Window *inWin) {

	// sanity check on input values
	// (if this fails here, it won't erase a previous successful call
	// since the prior init check comes second)
	if (inWin == nullptr) return 0;
	if (inWidth <= 0) return 0;
	if (inHeight <= 0) return 0;
	if (inWidth > SDL_VSCR_MAXWIDTH) return 0;
	if (inHeight > SDL_VSCR_MAXHEIGHT) return 0;

	// check to see if this was previously initialized
	if (this->win != nullptr) {
		// if so, reset our internal values to defaults
		this->width = this->height = 0;
		this->win = nullptr;
		if (this->tex != nullptr) SDL_DestroyTexture(this->tex);
		this->tex = nullptr;
		if (this->ren != nullptr) SDL_DestroyRenderer(this->ren);
		this->ren = nullptr;
	}


	this->width = inWidth;
	this->height = inHeight;

	this->win = inWin;
	this->ren = SDL_CreateRenderer(win, -1, SDL_RENDERER_ACCELERATED);
	if (this->ren == nullptr) return 0;

	this->tex = SDL_CreateTexture(ren, SDL_PIXELFORMAT_ARGB8888, SDL_TEXTUREACCESS_STREAMING, this->width, this->height);
	if (this->tex == nullptr) return 0;

	return 1;
}

int CSDLVirtualScreen::Update(unsigned char *inData) {
	if (this->win == nullptr || this->ren == nullptr || this->tex == nullptr) return 0;

	int y, x;
	int texW = 0;
	int texH = 0;
	int texAccess;
	Uint32 texFormat;
	Uint32 rMask, gMask, bMask, aMask;
	int rShift, gShift, bShift, aShift, bpp;
	Uint32 *texWritePtr;

	int pitch;
	char *pixels;
	char *offsetPtr;

	SDL_QueryTexture(this->tex, &texFormat, &texAccess, &texW, &texH);
	SDL_PixelFormatEnumToMasks(texFormat, &bpp, &rMask, &gMask, &bMask, &aMask);
	aShift = rShift = gShift = bShift = 0;
	if (aMask != 0x00000000) while (!(aMask & 0x00000001)) {
		aMask = aMask >> 1;
		aShift++;
	}
	if (rMask != 0x00000000) while (!(rMask & 0x00000001)) {
		rMask = rMask >> 1;
		rShift++;
	}
	if (gMask != 0x00000000) while (!(gMask & 0x00000001)) {
		gMask = gMask >> 1;
		gShift++;
	}
	if (bMask != 0x00000000) while (!(bMask & 0x00000001)) {
		bMask = bMask >> 1;
		bShift++;
	}
	

	SDL_LockTexture(this->tex, NULL, (void **)(&pixels), &pitch);

	for (y=0;y<texH;y++) {
		offsetPtr = pixels+(pitch*y);
		for (x=0;x<texW;x++) {
			texWritePtr = (Uint32 *)offsetPtr;
			*texWritePtr = 0x00000000;
			*texWritePtr = ((inData[0]&rMask)<<rShift) | ((inData[1]&gMask)<<gShift) | ((inData[2]&bMask)<<bShift) | ((0xFF&aMask)<<aShift);
			inData += 3;
			offsetPtr += 4;
		}
	}

	SDL_UnlockTexture(this->tex);

	return 1;
}

int CSDLVirtualScreen::Update(float *inData) {
	if (this->win == nullptr || this->ren == nullptr || this->tex == nullptr) return 0;


	int y, x;
	int texW = 0;
	int texH = 0;
	int texAccess;
	Uint32 texFormat;
	Uint32 rMask, gMask, bMask, aMask;
	int rShift, gShift, bShift, aShift, bpp;
	Uint32 *texWritePtr;

	int pitch;
	char *pixels;
	char *offsetPtr;

	Uint8	r, g, b;

	SDL_QueryTexture(this->tex, &texFormat, &texAccess, &texW, &texH);
	SDL_PixelFormatEnumToMasks(texFormat, &bpp, &rMask, &gMask, &bMask, &aMask);
	aShift = rShift = gShift = bShift = 0;
	if (aMask != 0x00000000) while (!(aMask & 0x00000001)) {
		aMask = aMask >> 1;
		aShift++;
	}
	if (rMask != 0x00000000) while (!(rMask & 0x00000001)) {
		rMask = rMask >> 1;
		rShift++;
	}
	if (gMask != 0x00000000) while (!(gMask & 0x00000001)) {
		gMask = gMask >> 1;
		gShift++;
	}
	if (bMask != 0x00000000) while (!(bMask & 0x00000001)) {
		bMask = bMask >> 1;
		bShift++;
	}
	

	SDL_LockTexture(this->tex, NULL, (void **)(&pixels), &pitch);

	for (y=0;y<texH;y++) {
		offsetPtr = pixels+(pitch*y);
		for (x=0;x<texW;x++) {
			texWritePtr = (Uint32 *)offsetPtr;
			*texWritePtr = 0x00000000;
			r = (Uint8)(*inData>1.0?255:(*inData)*255);
			inData++;
			g = (Uint8)(*inData>1.0?255:(*inData)*255);
			inData++;
			b = (Uint8)(*inData>1.0?255:(*inData)*255);
			inData++;
			*texWritePtr = ((r&rMask)<<rShift) | ((g&gMask)<<gShift) | ((b&bMask)<<bShift) | ((0xFF&aMask)<<aShift);
			offsetPtr += 4;
		}
	}

	SDL_UnlockTexture(this->tex);

	return 1;
}

int CSDLVirtualScreen::DirectUpdate(unsigned char *inData) {
	if (this->win == nullptr || this->ren == nullptr || this->tex == nullptr) return 0;

	int texW = 0;
	int texH = 0;
	int texAccess;
	Uint32 texFormat;

	int pitch;
	char *pixels;

	SDL_QueryTexture(this->tex, &texFormat, &texAccess, &texW, &texH);

	SDL_LockTexture(this->tex, NULL, (void **)(&pixels), &pitch);

	memcpy(pixels, inData, pitch*texH);

	SDL_UnlockTexture(this->tex);

	return 1;
}


void CSDLVirtualScreen::RenderTextureToScreen() {
	int screenW = 0;
	int screenH = 0;
	int texW = 0;
	int texH = 0;

	// additional storage for data returned by SDL_QueryTexture
	int texAccess;
	Uint32 texFormat;

	// A rectangle for scaling
	SDL_Rect outputRectangle;

	// get screen size
	SDL_GetWindowSize(this->win, &screenW, &screenH);

	// get texture size
	SDL_QueryTexture(this->tex, &texFormat, &texAccess, &texW, &texH);

	// assuming that we got good values back here ...
	if (((float)texW/(float)texH)>((float)screenW/(float)screenH)) {
		// texture wider than screen
		outputRectangle.w = screenW;
		outputRectangle.h = (texH*screenW)/texW;
		outputRectangle.x = 0;
		outputRectangle.y = (screenH-outputRectangle.h)>>1;
	} else {
		// texture taller than screen
		outputRectangle.h = screenH;
		outputRectangle.w = (texW*screenH)/texH;
		outputRectangle.y = 0;
		outputRectangle.x = (screenW-outputRectangle.w)>>1;
	}


	SDL_RenderClear(this->ren);
	SDL_RenderCopy(this->ren, this->tex, NULL, &outputRectangle);
	SDL_RenderPresent(this->ren);
}